# Copyright, 2015, by Samuel G. D. Williams. <http://www.codeotaku.com>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

require 'forwardable'

module Build
	module Dependency
		# Very similar to a set but uses a specific callback (defaults to &:name) for object identity.
		class Set
			include Enumerable
			
			def initialize(contents = [])
				@table = {}
				
				contents.each do |object|
					add(object)
				end
			end
			
			attr :table
			
			extend Forwardable
			
			def_delegators :@table, :size, :empty?, :clear, :count, :[], :to_s, :inspect
			
			def freeze
				return self if frozen?
				
				@table.freeze
				
				super
			end
			
			def initialize_dup(other)
				@table = other.table.dup
			end
			
			def identity(object)
				object.name
			end
			
			def add(object)
				if include?(object)
					raise KeyError, "Object #{identity(object)} already exists!"
				end
				
				@table[identity(object)] = object
			end
			
			alias << add
			
			def delete(object)
				@table.delete(identity(object))
			end
			
			def include?(object)
				@table.include?(identity(object))
			end
			
			def each(&block)
				@table.each_value(&block)
			end
			
			def slice(names)
				names.collect{|name| @table[name]}
			end
		end
	end
end
